@@ -25,7 +25,6 @@
 
 #include <errno.h>
 
-#define LIBUDEV_I_KNOW_THE_API_IS_SUBJECT_TO_CHANGE
 #include <libudev.h>
 
 #include <glib.h>
@@ -35,22 +34,6 @@
 #include <ofono/modem.h>
 #include <ofono/log.h>
 
-#ifdef NEED_UDEV_MONITOR_FILTER
-static int udev_monitor_filter_add_match_subsystem_devtype(struct udev_monitor *udev_monitor,
-				const char *subsystem, const char *devtype)
-{
-	return -EINVAL;
-}
-static int udev_monitor_filter_update(struct udev_monitor *udev_monitor)
-{
-	return -EINVAL;
-}
-static int udev_monitor_filter_remove(struct udev_monitor *udev_monitor)
-{
-	return -EINVAL;
-}
-#endif
-
 static GSList *modem_list = NULL;
 
 static struct ofono_modem *find_modem(const char *devpath)
@@ -94,32 +77,41 @@ static const char *get_driver(struct udev_device *udev_device)
 static void add_mbm(struct ofono_modem *modem,
 					struct udev_device *udev_device)
 {
-	const char *description, *devnode;
+	const char *desc, *devnode;
 	const char *device, *network;
 	int registered;
 
-	description = udev_device_get_sysattr_value(udev_device,
-							"device/interface");
-	if (description == NULL)
+	desc = udev_device_get_sysattr_value(udev_device, "device/interface");
+
+	if (desc == NULL)
 		return;
 
+	DBG("desc: %s", desc);
+
 	registered = ofono_modem_get_integer(modem, "Registered");
 	if (registered != 0)
 		return;
 
-	if (g_str_has_suffix(description, "Minicard Modem") == TRUE) {
+	if (g_str_has_suffix(desc, "Minicard Modem") ||
+			g_str_has_suffix(desc, "Mini-Card Modem") ||
+			g_str_has_suffix(desc, "Broadband Modem") ||
+			g_str_has_suffix(desc, "Broadband USB Modem")) {
 		devnode = udev_device_get_devnode(udev_device);
 		ofono_modem_set_string(modem, MODEM_DEVICE, devnode);
-	} else if (g_str_has_suffix(description,
-					"Minicard Data Modem") == TRUE) {
+	} else if (g_str_has_suffix(desc, "Minicard Data Modem") ||
+			g_str_has_suffix(desc, "Mini-Card Data Modem") ||
+			g_str_has_suffix(desc, "Broadband Data Modem")) {
 		devnode = udev_device_get_devnode(udev_device);
 		ofono_modem_set_string(modem, DATA_DEVICE, devnode);
-	} else if (g_str_has_suffix(description,
-					"Minicard GPS Port") == TRUE) {
+	} else if (g_str_has_suffix(desc, "Minicard GPS Port") ||
+			g_str_has_suffix(desc, "Mini-Card GPRS Port") ||
+			g_str_has_suffix(desc, "Broadband GPS Port")) {
 		devnode = udev_device_get_devnode(udev_device);
 		ofono_modem_set_string(modem, GPS_DEVICE, devnode);
-	} else if (g_str_has_suffix(description,
-					"Minicard Network Adapter") == TRUE) {
+	} else if (g_str_has_suffix(desc, "Minicard Network Adapter") ||
+			g_str_has_suffix(desc, "Mini-Card Network Adapter") ||
+			g_str_has_suffix(desc, "Broadband Network Adapter") ||
+			g_str_has_suffix(desc, "Minicard NetworkAdapter")) {
 		devnode = udev_device_get_property_value(udev_device,
 								"INTERFACE");
 		ofono_modem_set_string(modem, NETWORK_INTERFACE, devnode);
@@ -135,11 +127,14 @@ static void add_mbm(struct ofono_modem *modem,
 	}
 }
 
+#define APPLICATION_PORT "ApplicationPort"
+#define CONTROL_PORT "ControlPort"
+
 static void add_hso(struct ofono_modem *modem,
 					struct udev_device *udev_device)
 {
 	const char *subsystem, *type, *devnode;
-	const char *device, *network;
+	const char *app, *control, *network;
 	int registered;
 
 	subsystem = udev_device_get_subsystem(udev_device);
@@ -152,9 +147,13 @@ static void add_hso(struct ofono_modem *modem,
 
 	type = udev_device_get_sysattr_value(udev_device, "hsotype");
 
-	if (type != NULL && g_str_has_suffix(type, "Control") == TRUE) {
+	if (type != NULL) {
 		devnode = udev_device_get_devnode(udev_device);
-		ofono_modem_set_string(modem, MODEM_DEVICE, devnode);
+
+		if (g_str_has_suffix(type, "Application") == TRUE)
+			ofono_modem_set_string(modem, APPLICATION_PORT, devnode);
+		else if (g_str_has_suffix(type, "Control") == TRUE)
+			ofono_modem_set_string(modem, CONTROL_PORT, devnode);
 	} else if (g_str_equal(subsystem, "net") == TRUE) {
 		devnode = udev_device_get_property_value(udev_device,
 								"INTERFACE");
@@ -162,10 +161,11 @@ static void add_hso(struct ofono_modem *modem,
 	} else
 		return;
 
-	device  = ofono_modem_get_string(modem, MODEM_DEVICE);
+	app = ofono_modem_get_string(modem, APPLICATION_PORT);
+	control = ofono_modem_get_string(modem, CONTROL_PORT);
 	network = ofono_modem_get_string(modem, NETWORK_INTERFACE);
 
-	if (device != NULL && network != NULL) {
+	if (app != NULL && control != NULL && network != NULL) {
 		ofono_modem_set_integer(modem, "Registered", 1);
 		ofono_modem_register(modem);
 	}
